<IF><ARRAY>
$MSG "WARNING!;Array browsing is not implemented in this template"
<ENDIF>
$ACCEPT "What is your name?" TO AUTHOR
<IFKEY><EDIT>
<ELSE>
$ASKYN  "Hilite entire row?" TO HIROW
<ENDIF>
<IF><FREEZE>
$ASKYN "Lock frozen columns?" TO LOCK
<ENDIF>
$BROWSE aoBrowse[nBP]
$QUIT   lMore := .F.
$EDIT   DoGet(aoBrowse[nBP]) ; aoBrowse[nBP]:refreshCurrent()
$SWITCH aoBrowse[nBP]:deHilite(); IF(nBP < LEN(aoBrowse), nBP++, nBP := 1); DispBrowse(aoBrowse[nBP])
$MOVE   MoveBrowse(aoBrowse[nBP]); DrawAll(aoBrowse, nBP)
$SIZE   SizeBrowse(aoBrowse[nBP]); DrawAll(aoBrowse, nBP)
$DATA   aData
/* ..........................................................................
 * 
 *    Program Name: <FILENAME>
 *          Author: <AUTHOR>
 *    Date Created: <DATE>
 *        Language: Clipper 5.01
 * 
 * Notes  : Multiple browse windows support
 *	       : Compile with /n/w/m
 * ..........................................................................
 */

#include "setcurs.ch"
<IFKEY><SIZE><OR><MOVE>
#include "inkey.ch"
<ENDIF>

#define ALIAS	  1
#define BOXTOP    2
#define BOXLEFT   3
#define BOXBOTTOM 4
#define BOXRIGHT  5
#define BOXTYPE   6
#define BOXBORDER 7
#define BOXINCOLOR 8

//===========================================================================
/*
 BrowseIt()

 Opens appropriate database and index files, sets relations,
 sets up the columns, and browses the data.
 Restores user screen upon exit.

 SYNTAX:  BrowseIt()

 RETURNS: NIL
 --------------------------------------------------------------------------*/
FUNCTION BrowseIt()
	LOCAL aoBrowse := ARRAY(<BROWSECNT>)
	LOCAL nCursSave          // saves original cursor
	LOCAL cSaveScr, nSaveRow // saves original user screen
	LOCAL lMore              // controls processing
	LOCAL nKey               // receives keypress from user
	LOCAL nBP                // current browse pointer into array
	LOCAL n                  // general index counter

	SET SCOREBOARD OFF

	// save original user screen so we can restore upon exit
	cSaveScr := SAVESCREEN(0, 0, MAXROW(), MAXCOL())
        nSaveRow := MAXROW()

	// open database files
	<ALLDBF>
	DBUSEAREA(.T.,, "<DBF>")
	DBSETINDEX("<NTX>")
	<ENDDBF>

	// start screen buffering while we display screen
	DISPBEGIN()

	CLEAR SCREEN
	SETMODE(<MAXROW>, 80)

	<ALLBROWSE>
	// set up browse and display it
	aoBrowse[<BROWSENUM>] := <PREFIX><BROWSENUM>Browse()
	aoBrowse[<BROWSENUM>]:goTop()
	DispBrowse(aoBrowse[<BROWSENUM>])

	<ENDBROWSE>
	DISPEND()                // end display buffering

	// turn cursor off while browsing
	nCursSave := SETCURSOR(SC_NONE)

	// set browse pointer to first browse defined
	nBP := LEN(aoBrowse)
	DispBrowse(aoBrowse[nBP])

	// main browse display loop
	lMore := .T.
	DO WHILE lMore
		<IF><LOCK>
		// lock frozen columns
		IF aoBrowse[nBP]:colPos <= aoBrowse[nBP]:freeze
			aoBrowse[nBP]:colPos := aoBrowse[nBP]:freeze + 1
		ENDIF

		<ENDIF>
		nKey := 0
		// Stabilize the browse until it's stable or a key is pressed
		DO WHILE (nKey := INKEY()) == 0 .AND. !aoBrowse[nBP]:stabilize()
		ENDDO

		IF aoBrowse[nBP]:stable

			// alert if hit top or bottom of browse
			IF aoBrowse[nBP]:hitTop .OR. aoBrowse[nBP]:hitBottom
				TONE(125, 0)
			ENDIF

			<IF><RELATION>
			// update other browse screens in case relations are set
			IF !EMPTY(DBRELATION(1))
				DrawAll(aoBrowse, nBP)
			ENDIF
			<ENDIF>

			<IF><HIROW>
			aoBrowse[nBP]:colorRect({aoBrowse[nBP]:rowPos, 1, aoBrowse[nBP]:rowPos, aoBrowse[nBP]:colCount}, {2, 1})
			<ELSE>
			aoBrowse[nBP]:hilite()
			<ENDIF>

			// Get a new keystroke
			nKey := INKEY(0)

			<IF><HIROW>
			aoBrowse[nBP]:refreshCurrent()
			<ENDIF>

		ENDIF


		// Apply the key to the browse
		lMore := ApplyKey(aoBrowse, nKey, @nBP)

	ENDDO

	// restore original cursor
	SETCURSOR(nCursSave)

	// restore original user screen
	SETMODE(nSaveRow, 80)
	RESTSCREEN(0, 0, MAXROW(), MAXCOL(), cSaveScr)
	SETPOS(nSaveRow, 0)

	RETURN (NIL)


//===========================================================================
/*
 ForceStable()

 Stabilizes a browse object

 SYNTAX:  ForceStable(expO1)

    <expO1> -  browse object to stabilize
 --------------------------------------------------------------------------*/
PROCEDURE ForceStable( oBrowse )

	DISPBEGIN()
	DO WHILE .NOT. oBrowse:stabilize()
	ENDDO
	DISPEND()

	RETURN


FUNCTION ApplyKey(aoBrowse, nKey, nBP)
	LOCAL lMore := .T.

	DO CASE
	<ALLKEYS>
	CASE nKey == <INKEY>	// <KEY>
		<ACTION>
	<ENDKEYS>

	ENDCASE

	RETURN lMore

<IF><PICTURE>

//===========================================================================
/*
 PictBlock()

 Sets up a column to handle display and editing of picture codes.

 SYNTAX:  PictBlock(expC1, expC2, expO1)

    <expC1> -  name of field to add to column
    <expC2> -  picture clause for the field
    <expO1> -  column object to add picture code to

 RETURNS: the modified column object
 --------------------------------------------------------------------------*/
FUNCTION PictBlock(cField, cPicture, oCol)
	oCol:block := &('{|xVal| IF(xVal == NIL, TRANSFORM(' ;
		+cField+ ', "'+cPicture+'"), '+cField+' := xVal ) }')
	oCol:cargo := {FIELDBLOCK(cField), cPicture}

	RETURN oCol
<ENDIF>
<IF><FILTER>

//===========================================================================
/*
 GoNext()

 Skip routine for a filtered browse

 SYNTAX:  GoNext(expN1, expB1)

    <expN1> -  number of records to skip
    <expB1> -  block while condition to test for during skip

 RETURNS: number of records actually skipped
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoNext(nToSkip, bWhileCond )
	LOCAL nSkipped := 0, nDirection

	nDirection := IIF(nToSkip > 0, 1, -1)

	DO WHILE nSkipped != nToSkip .AND. EVAL(bWhileCond) .AND. ;
	  !EOF() .AND. !BOF()
		SKIP nDirection
		nSkipped += nDirection
	ENDDO

	IF EOF()
		SKIP -1
		nSkipped--
	ELSEIF BOF()
 		GOTO RECNO()
		nSkipped++
	ELSEIF !EVAL(bWhileCond)
		SKIP -nDirection
		nSkipped += -nDirection
	ENDIF

	RETURN nSkipped


//===========================================================================
/*
 GoTop()

 Goes to top of data in a filtered browse.

 SYNTAX:  GoTop(expB1)

    <expB1> -  block while condition to test for

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoTop( bWhileCond )
   WHILE EVAL( bWhileCond ) .AND. !BOF()
      SKIP -1

      IF !EVAL( bWhileCond )
         SKIP
         EXIT
      ENDIF

   END
   RETURN (NIL)


//===========================================================================
/*
 GoBottom()

 Goes to bottom of data in a filtered browse.

 SYNTAX:  GoBottom(expB1)

    <expB1> -  block while condition to test for

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoBottom( bWhileCond )
	WHILE EVAL( bWhileCond ) .AND. !EOF()
		SKIP
	END

	// We always skip 1 too far, so skip back
	SKIP -1

	RETURN (NIL)
<ENDIF>

<ALLBROWSE>
// set up <DBFFILE> database
FUNCTION <PREFIX><BROWSENUM>Browse(oBrowse)
	LOCAL oCol

	SELECT <ALIAS>
	<IF><FILTER>

	// If you know your filter condition is an equality statement on
	// ...the index key, use DBSEEK(<key>) for better performance
	LOCATE FOR <FILTER>

	<ENDIF>
	SET RELATION TO <RELATIONS>

	oBrowse := TBrowseDB( <BROWSETOP>, <BROWSELEFT>, <BROWSEBOTTOM>, <BROWSERIGHT>)
	<IF><ARRAY>
	oBrowse:skipBlock := {|nSkip| -np + (np := MAX(1, MIN(LEN(<DATA>), np + nSkip))) }
	oBrowse:goTopBlock := {|| np := 1}
	oBrowse:goBottomBlock := {|| np := LEN(<DATA>) }
	<ELSE>
	<IF><FILTER>
	oBrowse:skipBlock     := {|x| GoNext(x, {|| <FILTER> } ) }
	oBrowse:goTopBlock    := {|| GoTop( {|| <FILTER> } ) }
	oBrowse:goBottomBlock := {|| GoBottom( {|| <FILTER> } ) }
	<ENDIF>
	<ENDIF>
	// define default column attributes
	oBrowse:colSep  := <COLSEP>
	oBrowse:headSep := <HEADSEP>
	oBrowse:footSep := <FOOTSEP>
	oBrowse:colorSpec := <COLORSPEC>
	oBrowse:autoLite := .F.

	// store the work area in the cargo so we can do multiple browses
	oBrowse:cargo := {SELECT(), <BOXTOP>, <BOXLEFT>, ;
		<BOXBOTTOM>, <BOXRIGHT>, <BOXTYPE>, ;
		<INSIDECOLOR>, <OUTLINECOLOR>}

	<ALLCOLUMNS>
	<IF><ARRAY>
	<IF><PICTURE>
	oCol := TBColumnNew(<HEADING>, {|| TRANSFORM(<DATA>[np][<ARRAYNAME>], "<PICTURE>") })
	<ELSE>
	oCol := TBColumnNew(<HEADING>, {|| <DATA>[np][<ARRAYNAME>] })
	<ENDIF>
	<ELSE>
	<IF><FIELD>
	<IF><PICTURE>
	oCol := TBColumnNew(<HEADING>, NIL)
	<IFKEY><EDIT>
	<ENDIF>
	oCol := PictBlock("<CONTENTS>", "<PICTURE>", oCol)
	<ELSE>
	oCol := TBColumnNew(<HEADING>, FIELDBLOCK("<CONTENTS>") )
	<IFKEY><EDIT>
	oCol:cargo := {FIELDBLOCK("<CONTENTS>"), ""}
	<ENDIF>
	<ENDIF>
	<ELSE>
	oCol := TBColumnNew(<HEADING>, {|| <CONTENTS> } )
	<ENDIF>
	<ENDIF>
	oCol:headSep	:= <COLHEADSEP>
	oCol:footSep	:= <COLFOOTSEP>
	oCol:footing	:= <FOOTING>
	oCol:width	:= <COLWIDTH>
	oCol:colSep	:= <COLCOLSEP>
	oCol:defColor	:= <COLCOLOR>
	oCol:colorBlock	:= {|xVal| <COLORBLOCK> }
	oBrowse:addColumn(oCol)

	<ENDCOLUMNS>
	oBrowse:freeze := <FREEZE>

	RETURN oBrowse

<ENDBROWSE>

<IF><RELATION>
FUNCTION DrawAll(aoBrowse, nBP)
	LOCAL n

	SETCOLOR("W/N")
	DISPBEGIN()

	CLEAR SCREEN
	FOR n := 1 TO LEN(aoBrowse)
		IF n != nBP
			DispBrowse(aoBrowse[n])
		ENDIF
	NEXT
	DispBrowse(aoBrowse[nBP])

	DISPEND()

	RETURN (NIL)

<ELSE>
<IFKEY><MOVE><OR><SIZE>
FUNCTION DrawAll(aoBrowse, nBP)
	LOCAL n

	DISPBEGIN()

	CLEAR SCREEN
	FOR n := 1 TO LEN(aoBrowse)
		IF n != nBP
			DispBrowse(aoBrowse[n])
		ENDIF
	NEXT
	DispBrowse(aoBrowse[nBP])

	DISPEND()

	RETURN (NIL)

<ENDIF>
<ENDIF>

// Display the browse screen
FUNCTION DispBrowse(oBrowse)

	// select the database for this browse
	SELECT (oBrowse:cargo[ALIAS])

	DISPBEGIN()

	// clear old screen area
	@ oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT] CLEAR TO oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT]

	// display the browse box
	DISPBOX(oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT],;
		oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT],;
		oBrowse:cargo[BOXTYPE], oBrowse:cargo[BOXBORDER])

	// refresh browse so all data is redisplayed in the box
	oBrowse:refreshAll()
	ForceStable(oBrowse)

	DISPEND()

	RETURN (NIL)

<IFKEY><EDIT>

STATIC FUNCTION ColWidth(oCol)
	LOCAL xValue
	LOCAL nWidth

	nWidth := oCol:width
	IF nWidth == NIL
		xValue := EVAL(oCol:block)
		DO CASE
		CASE VALTYPE(xValue) == "C"
			nWidth := LEN(xValue)
		CASE VALTYPE(xValue) == "N"
			nWidth := LEN(STR(xValue))
		CASE VALTYPE(xValue) == "D"
			nWidth := LEN(DTOC(xValue))
		CASE VALTYPE(xValue) == "L"
			nWidth := 1
		ENDCASE
		IF oCol:Heading != NIL
			nWidth := MAX(nWidth, LEN(oCol:heading) )
		ENDIF
	ENDIF

	RETURN nWidth

<ENDIF>

<IFKEY><EDIT>
//===========================================================================
/*
 DoGet()

 Edits current cell using the appropriate picture codes for the column

 SYNTAX:  DoGet(expO1)

    <expO1> -  current browse object

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION DoGet(oBrowse)
	LOCAL oCol
	LOCAL xValue
	LOCAL nWidth
	LOCAL xSaveKey
	LOCAL nRecNo

	oCol := oBrowse:getColumn(oBrowse:colPos)

	<IF><RELATION>
	// make sure our relation gets set to the right record!
	GOTO RECNO()
	<ENDIF>

	// save controlling index key value
	// save controlling index key value
	IF !EMPTY(INDEXKEY(0))
		xSaveKey := &(INDEXKEY(0))
		nRecNo   := RECNO()
	ENDIF

	// get ACTUAL width of column so we don't overwrite adjacent columns
	nWidth := ColWidth(oCol)

	// turn cursor on and do read at current position
	SETCURSOR(SC_NORMAL)
	READMODAL({GetNew(ROW(), COL(), oCol:cargo[1],, ;
	  "@S"+LTRIM(STR(nWidth,2))+oCol:cargo[2], oBrowse:colorSpec)})

	// turn cursor off
	SETCURSOR(SC_NONE)


	// if index key has changed, refresh the browse
	IF !EMPTY(xSaveKey) .AND. !(&(INDEXKEY(0)) == xSaveKey )
		oBrowse:refreshAll()
		ForceStable( oBrowse )
		DO WHILE (RECNO() != nRecNo)
			oBrowse:up()
			ForceStable( oBrowse )
		ENDDO
	ENDIF


	RETURN NIL

<ENDIF>
<IFKEY><MOVE>

//===========================================================================
/*
 MoveBrowse()

 Dynamically move the browse box.
 Use the arrow keys to move the box, <return> to select position.
 Updates browse object to reflect new box coordinates.

 SYNTAX:  MoveBrowse(expO1)

    <expB1> -  browse object we are moving

 RETURNS: NIL
 --------------------------------------------------------------------------*/
FUNCTION MoveBrowse( oBrowse )
	LOCAL cSaveScr
	LOCAL cColor
	LOCAL nKey
	LOCAL nTop, nLeft, nBottom, nRight

	nTop := oBrowse:cargo[BOXTOP]
	nLeft := oBrowse:cargo[BOXLEFT]
	nBottom := oBrowse:cargo[BOXBOTTOM]
	nRight := oBrowse:cargo[BOXRIGHT]

	// save original screen
	cColor := SETCOLOR("W+/N")
	cSaveScr  := SAVESCREEN( 0, 0, MAXROW(), MAXCOL() )

	@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	DO WHILE .T.
		nKey := INKEY(0)
		RESTSCREEN( 0, 0, MAXROW(), MAXCOL(), cSaveScr )

		DO CASE
		CASE nKey == K_UP
			IF ( nTop != 0 )
				nTop--
				nBottom--
			ENDIF
		CASE nKey == K_DOWN
			IF nBottom != MAXROW()
				nBottom++
				nTop++
			ENDIF
		CASE nKey == K_LEFT
			IF nLeft != 0
				nLeft--
				nRight--
			ENDIF
		CASE nKey == K_RIGHT
			IF nRight != MAXCOL()
				nRight++
				nLeft++
			ENDIF
		CASE nKey == K_ENTER
			EXIT
		ENDCASE
		@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	ENDDO
	SETCOLOR(cColor)

	oBrowse:nTop += (nTop - oBrowse:cargo[BOXTOP])
	oBrowse:nLeft += (nLeft - oBrowse:cargo[BOXLEFT])
	oBrowse:nBottom += (nBottom - oBrowse:cargo[BOXBOTTOM])
	oBrowse:nRight  += (nRight - oBrowse:cargo[BOXRIGHT])

	oBrowse:cargo[BOXTOP] := nTop
	oBrowse:cargo[BOXLEFT] := nLeft
	oBrowse:cargo[BOXBOTTOM] := nBottom
	oBrowse:cargo[BOXRIGHT] := nRight

	RETURN (NIL)

<ENDIF>
<IFKEY><SIZE>

//===========================================================================
/*
 SizeBrowse()

 Dynamically size the browse box.
 Use the arrow keys to size the box, <return> to select size.
 Updates browse object to reflect new box coordinates.

 SYNTAX:  SizeBrowse(expO1)

    <expB1> -  browse object we are sizing

 RETURNS: NIL
 --------------------------------------------------------------------------*/
FUNCTION SizeBrowse( oBrowse )
	LOCAL cSaveScr
	LOCAL cColor
	LOCAL nKey
	LOCAL nTop, nLeft, nBottom, nRight

	nTop := oBrowse:cargo[BOXTOP]
	nLeft := oBrowse:cargo[BOXLEFT]
	nBottom := oBrowse:cargo[BOXBOTTOM]
	nRight := oBrowse:cargo[BOXRIGHT]

	// save original screen
	cColor := SETCOLOR("W+/N")
	cSaveScr := SAVESCREEN( 0, 0, MAXROW(), MAXCOL() )

	@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	DO WHILE .T.
		nKey := INKEY(0)
		RESTSCREEN( 0, 0, MAXROW(), MAXCOL(), cSaveScr )

		DO CASE
		CASE nKey == K_UP
			nBottom := IF( nBottom > nTop+2, --nBottom, nBottom )
		CASE nKey == K_DOWN
			nBottom := IF( nBottom != MAXROW(), ++nBottom, nBottom )
		CASE nKey == K_LEFT
			nRight := IF( nRight > nLeft+2, --nRight, nRight )
		CASE nKey == K_RIGHT
			nRight := IF( nRight != MAXCOL(), ++nRight, nRight )
		CASE nKey == K_ENTER
			EXIT
		ENDCASE

		@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	ENDDO
	SETCOLOR(cColor)


	oBrowse:nBottom += (nBottom - oBrowse:cargo[BOXBOTTOM])
	oBrowse:nRight  += (nRight - oBrowse:cargo[BOXRIGHT])
	oBrowse:cargo[BOXBOTTOM] := nBottom
	oBrowse:cargo[BOXRIGHT] := nRight


	RETURN (NIL)

<ENDIF>

