$ACCEPT "What is your name?" TO AUTHOR
<IFKEY><EDIT>
<ELSE>
$ASKYN  "Hilite entire row?" TO HIROW
<ENDIF>
<IF><FREEZE>
$ASKYN  "Lock frozen columns?" TO LOCK
<ENDIF>
$BROWSE oBrowse
$DATA   aData
$QUIT   lMore := .F.
$EDIT   DoGet(oBrowse) ; oBrowse:refreshCurrent()
$MOVE   MoveBrowse(oBrowse); ForceStable(oBrowse)
$SIZE   SizeBrowse(oBrowse); ForceStable(oBrowse)
/* ..........................................................................
 * 
 *    Program Name: <FILENAME>
 *          Author: <AUTHOR>
 *    Date Created: <DATE>
 *        Language: Clipper 5.01
 * 
 * Notes  : Single browse window only
 *	       : Compile with /n/w/m
 * ..........................................................................
 */

#include "setcurs.ch"
<IFKEY><SIZE><OR><MOVE>
#include "inkey.ch"
<ENDIF>

<IFKEY><SIZE><OR><MOVE>
#define ALIAS	  1
#define BOXTOP    2
#define BOXLEFT   3
#define BOXBOTTOM 4
#define BOXRIGHT  5
#define BOXTYPE   6
#define BOXBORDER 7
#define BOXINCOLOR 8

<ENDIF>
<IF><ARRAY>
$ASKYN   "Generate test code?" TO TEST
<ALLARRAY>
#define	<ARRAYNAME>	<ARRAYPOS>
<ENDARRAY>
<IF><TEST>

// Delete or comment out the following procedure when finished testing!
PROCEDURE TEST()
	LOCAL aTest := {}
	
	FOR n := 1 TO 55
		AADD(aTest, {})
		<ALLARRAY>
		AADD(aTest[n], <ARRAYSAMPLE>)
		<ENDARRAY>
	NEXT

	BrowseIt(aTest)

	RETURN

<ENDIF>
<ENDIF>

//===========================================================================
/*
 BrowseIt()

 Opens appropriate database and index files, sets relations,
 sets up the columns, and browses the data.
 Restores user screen upon exit.

<IF><ARRAY>
 SYNTAX:  BrowseIt(expA1)

    <expA1> -  array of column data to browse

 RETURNS: Index into current array element
<ELSE>
 SYNTAX:  BrowseIt()

 RETURNS: NIL
<ENDIF>
 --------------------------------------------------------------------------*/
<IF><ARRAY>
FUNCTION BrowseIt(<DATA>)
<ELSE>
FUNCTION BrowseIt()
<ENDIF>
	
	LOCAL oBrowse, oCol      // browse & column objects
	LOCAL nCursSave          // stores state of cursor
	LOCAL lMore              // controls processing
	LOCAL nKey               // receives keypress
	LOCAL n                  // index counter
	LOCAL cSaveScr, nSaveRow // saves original user screen
	<IF><ARRAY>
	LOCAL np := 1            // index pointer into array
	<ENDIF>

	SET SCOREBOARD OFF
	
	// save original user screen so we can restore upon exit
	cSaveScr := SAVESCREEN(0, 0, MAXROW(), MAXCOL())
        nSaveRow := MAXROW()

	// start screen buffering while we display screen
	DISPBEGIN()

	CLEAR SCREEN
	SETMODE(<MAXROW>, 80)

	<ALLBROWSE>
	// Display browse box
	DISPBOX( <BOXTOP>, <BOXLEFT>, <BOXBOTTOM>, <BOXRIGHT>, "         ", <INSIDECOLOR>)
	DISPBOX( <BOXTOP>, <BOXLEFT>, <BOXBOTTOM>, <BOXRIGHT>, <BOXTYPE>, <OUTLINECOLOR>)
	<ENDBROWSE>

	DISPEND()                // end display buffering

	// open database files
	<ALLDBF>
	DBUSEAREA(.T.,, "<DBF>")
	DBSETINDEX("<NTX>")
	<ENDDBF>

	<ALLBROWSE>
	<IF><ARRAY>
	<ELSE>
	SELECT("<ALIAS>")
	<ENDIF>
	<IF><FILTER>

	// If you know your filter condition is an equality statement on
	// ...the index key, use DBSEEK(<key>) for better performance
	LOCATE FOR <FILTER>

	<ENDIF>
	SET RELATION TO <RELATIONS>
	
	<ENDBROWSE>
	// set up browse and display it
	<ALLBROWSE>
	oBrowse := TBrowseDB( <BROWSETOP>, <BROWSELEFT>, <BROWSEBOTTOM>, <BROWSERIGHT>)
	<IF><ARRAY>
	oBrowse:skipBlock := {|nSkip| -np + (np := MAX(1, MIN(LEN(<DATA>), np + nSkip))) }
	oBrowse:goTopBlock := {|| np := 1}
	oBrowse:goBottomBlock := {|| np := LEN(<DATA>) }
	<ELSE>
	<IF><FILTER>
	oBrowse:skipBlock     := {|x| GoNext(x, {|| <FILTER> } ) }
	oBrowse:goTopBlock    := {|| GoTop( <FILTER> ) }
	oBrowse:goBottomBlock := {|| GoBottom( {|| <FILTER> } ) }

	<ENDIF>
	<ENDIF>
	// define default column attributes
	oBrowse:colSep  := <COLSEP>
	oBrowse:headSep := <HEADSEP>
	oBrowse:footSep := <FOOTSEP>
	oBrowse:colorSpec := <COLORSPEC>
	<IF><HIROW>
	oBrowse:autoLite := .F.
	<ENDIF>
	<IFKEY><SIZE><OR><MOVE>

	// store box attributes so we can move & size the box
	// we store the work area just to maintain consistency with MULTI.BRO
	oBrowse:cargo := {SELECT(), <BOXTOP>, <BOXLEFT>, ;
		<BOXBOTTOM>, <BOXRIGHT>, <BOXTYPE>, ;
		<INSIDECOLOR>, <OUTLINECOLOR>}
	<ENDIF>

	// set up all columns
	<ALLCOLUMNS>
	<IF><ARRAY>
	oCol := TBColumnNew(<HEADING>, NIL)
	<IF><PICTURE>
	oCol:block := {|xVal| IF(xVal == NIL, TRANSFORM(<DATA>[np][<ARRAYNAME>], "<PICTURE>"), <DATA>[np][<ARRAYNAME>] := xVal) }
	<ELSE>
	oCol:block := {|xVal| IF(xVal == NIL, <DATA>[np][<ARRAYNAME>], <DATA>[np][<ARRAYNAME>] := xVal) }
	<ENDIF>
	<IFKEY><EDIT>
	<IF><PICTURE>
	oCol:cargo := "<PICTURE>"
	<ELSE>
	oCol:cargo := ""
	<ENDIF>
	<ENDIF>
	<ELSE>
	<IF><FIELD>
	<IF><PICTURE>
	oCol := TBColumnNew(<HEADING>, NIL)
	oCol := PictBlock("<CONTENTS>", "<PICTURE>", oCol)
	<ELSE>
	oCol := TBColumnNew(<HEADING>, FIELDWBLOCK("<FIELD>", SELECT("<FLDALIAS>")) )
	<IFKEY><EDIT>
	// store the information for cell editing in the column cargo 
	oCol:cargo := {FIELDWBLOCK("<CONTENTS>", SELECT("<FLDALIAS>")), ""}
	<ENDIF>
	<ENDIF>
	<ELSE>
	<IF><PICTURE>
	oCol := TBColumnNew(<HEADING>, {|| TRANSFORM(<CONTENTS>, "<PICTURE>") } )
	<IFKEY><EDIT>
	// cargo contains the picture clause for this column
	oCol:cargo := {FIELDBLOCK("<CONTENTS>"), "<PICTURE>"}
	<ENDIF>
	<ELSE>
	oCol := TBColumnNew(<HEADING>, {|| <CONTENTS> } )
	<IFKEY><EDIT>
	// cargo contains the picture clause for this column
	oCol:cargo := {FIELDBLOCK("<CONTENTS>"), ""}
	<ENDIF>
	<ENDIF>
	<ENDIF>
	<ENDIF>
	oCol:headSep	:= <COLHEADSEP>
	oCol:footSep	:= <COLFOOTSEP>
	oCol:footing	:= <FOOTING>
	oCol:width	:= <COLWIDTH>
	oCol:colSep	:= <COLCOLSEP>
	oCol:defColor	:= <COLCOLOR>
	oCol:colorBlock	:= {|xVal| <COLORBLOCK> }
	oBrowse:addColumn(oCol)

	<ENDCOLUMNS>
	oBrowse:freeze := <FREEZE>
	<ENDBROWSE>

	// turn cursor off while browsing
	nCursSave := SETCURSOR(SC_NONE)

	// main browse display loop
	lMore := .T.
	DO WHILE lMore

		<IF><LOCK>
		// lock frozen columns
		IF oBrowse:colPos <= oBrowse:freeze
			oBrowse:colPos := oBrowse:freeze + 1
		ENDIF

		<ENDIF>
		nKey := 0

		// Stabilize the browse until it's stable or a key is pressed
		DO WHILE (nKey := INKEY()) == 0 .AND. !oBrowse:stabilize()
		ENDDO

		IF oBrowse:stable

			// alert if hit top or bottom of browse
			IF oBrowse:hitTop .OR. oBrowse:hitBottom
				TONE(125, 0)
			ENDIF

			<IF><HIROW>
			// color entire row, not just a single cell
			oBrowse:colorRect({oBrowse:rowPos, 1, oBrowse:rowPos, oBrowse:colCount}, {2, 1})
			<ENDIF>

			// everything updated, now wait for a keypress
			nKey := INKEY(0)
		ENDIF

		<IF><HIROW>
		// de-hilite the entire row the next time we update browse
		oBrowse:refreshCurrent()
		<ENDIF>

		// Apply the key to the browse
		DO CASE
		<ALLKEYS>
		CASE nKey == <INKEY>	// <KEY>
			<ACTION>
		<ENDKEYS>

		ENDCASE

	ENDDO

	// restore original cursor
	SETCURSOR(nCursSave)

	// restore original user screen
	SETMODE(nSaveRow, 80)
	RESTSCREEN(0, 0, MAXROW(), MAXCOL(), cSaveScr)
	SETPOS(nSaveRow, 0)

	<IF><ARRAY>
	// return the current array element
	RETURN np
	<ELSE>
	RETURN (NIL)
	<ENDIF>

<IF><PICTURE>

//===========================================================================
/*
 PictBlock()

 Sets up a column to handle display and editing of picture codes.

 SYNTAX:  PictBlock(expC1, expC2, expO1)

    <expC1> -  name of field to add to column
    <expC2> -  picture clause for the field
    <expO1> -  column object to add picture code to

 RETURNS: the modified column object
 --------------------------------------------------------------------------*/
FUNCTION PictBlock(cField, cPicture, oCol)
	oCol:block := &('{|xVal| IF(xVal == NIL, TRANSFORM(' ;
		+cField+ ', "'+cPicture+'"), '+cField+' := xVal ) }')
	oCol:cargo := {FIELDBLOCK(cField), cPicture}

	RETURN oCol

<ENDIF>
<IFKEY><EDIT>

//===========================================================================
/*
 ForceStable()

 Stabilizes a browse object

 SYNTAX:  ForceStable(expO1)

    <expO1> -  browse object to stabilize
 --------------------------------------------------------------------------*/
PROCEDURE ForceStable( oBrowse )

	DISPBEGIN()
	DO WHILE .NOT. oBrowse:stabilize()
	ENDDO
	DISPEND()

	RETURN


//===========================================================================
/*
 DoGet()

 Edits current cell using the appropriate picture codes for the column

 SYNTAX:  DoGet(expO1)

    <expO1> -  current browse object

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION DoGet(oBrowse)
	LOCAL oCol
	LOCAL xValue
	LOCAL nWidth
	LOCAL nRecNo
	LOCAL xSaveKey

	oCol := oBrowse:getColumn(oBrowse:colPos)

	// get ACTUAL width of column so we don't overwrite adjacent columns
	nWidth := oCol:width
	IF nWidth == NIL
		xValue := EVAL(oCol:block)
		DO CASE
		CASE VALTYPE(xValue) == "C"
			nWidth := LEN(xValue)
		CASE VALTYPE(xValue) == "N"
			nWidth := LEN(STR(xValue))
		CASE VALTYPE(xValue) == "D"
			nWidth := LEN(DTOC(xValue))
		CASE VALTYPE(xValue) == "L"
			nWidth := 1
		ENDCASE
		IF oCol:Heading != NIL
			nWidth := MAX(nWidth, LEN(oCol:heading) )
		ENDIF
	ENDIF
	
	// save controlling index key value
	IF !EMPTY(INDEXKEY(0))
		xSaveKey := &(INDEXKEY(0))
		nRecNo   := RECNO()
	ENDIF


	// turn cursor on and do read at current position
	SETCURSOR(SC_NORMAL)
	READMODAL({GetNew(ROW(), COL(), oCol:cargo[1],, ;
	  "@S"+LTRIM(STR(nWidth,2))+oCol:cargo[2], oBrowse:colorSpec)})

	// turn cursor off
	SETCURSOR(SC_NONE)


	IF !EMPTY(xSaveKey) .AND. !(&(INDEXKEY(0)) == xSaveKey )
		oBrowse:refreshAll()
		ForceStable( oBrowse )
		DO WHILE (RECNO() != nRecNo)
			oBrowse:up()
			ForceStable( oBrowse )
		ENDDO
	ENDIF

	RETURN (NIL)

<ENDIF>
<IF><FILTER>

//===========================================================================
/*
 GoNext()

 Skip routine for a filtered browse

 SYNTAX:  GoNext(expN1, expB1)

    <expN1> -  number of records to skip
    <expB1> -  block while condition to test for during skip

 RETURNS: number of records actually skipped
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoNext(nToSkip, bWhileCond )
	LOCAL nSkipped := 0, nDirection

	nDirection := IIF(nToSkip > 0, 1, -1)

	DO WHILE nSkipped != nToSkip .AND. EVAL(bWhileCond) .AND. ;
	  !EOF() .AND. !BOF()
		SKIP nDirection
		nSkipped += nDirection
	ENDDO

	IF EOF()
		SKIP -1
		nSkipped--
	ELSEIF BOF()
 		GOTO RECNO()
		nSkipped++
	ELSEIF !EVAL(bWhileCond)
		SKIP -nDirection
		nSkipped += -nDirection
	ENDIF

	RETURN nSkipped


//===========================================================================
/*
 GoTop()

 Goes to top of data in a filtered browse.

 SYNTAX:  GoTop(expB1)

    <expB1> -  block while condition to test for

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoTop( bWhileCond )
   WHILE EVAL( bWhileCond ) .AND. !BOF()
      SKIP -1

      IF !EVAL( bWhileCond )
         SKIP
         EXIT
      ENDIF

   END
   RETURN (NIL)


//===========================================================================
/*
 GoBottom()

 Goes to bottom of data in a filtered browse.

 SYNTAX:  GoBottom(expB1)

    <expB1> -  block while condition to test for

 RETURNS: NIL
 --------------------------------------------------------------------------*/
STATIC FUNCTION GoBottom( bWhileCond )
	WHILE EVAL( bWhileCond ) .AND. !EOF()
		SKIP
	END

	// We always skip 1 too far, so skip back
	SKIP -1

	RETURN (NIL)

<ENDIF>
<IFKEY><MOVE>

//===========================================================================
/*
 MoveBrowse()

 Dynamically move the browse box.
 Use the arrow keys to move the box, <return> to select position.
 Updates browse object to reflect new box coordinates.

 SYNTAX:  MoveBrowse(expO1)

    <expB1> -  browse object we are moving

 RETURNS: NIL
 --------------------------------------------------------------------------*/
FUNCTION MoveBrowse( oBrowse )
	LOCAL cSaveScr
	LOCAL cColor
	LOCAL nKey
	LOCAL nTop, nLeft, nBottom, nRight

	nTop := oBrowse:cargo[BOXTOP]
	nLeft := oBrowse:cargo[BOXLEFT]
	nBottom := oBrowse:cargo[BOXBOTTOM]
	nRight := oBrowse:cargo[BOXRIGHT]

	// save original screen
	cColor := SETCOLOR("W+/N")
	cSaveScr  := SAVESCREEN( 0, 0, MAXROW(), MAXCOL() )

	@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	DO WHILE .T.
		nKey := INKEY(0)
		RESTSCREEN( 0, 0, MAXROW(), MAXCOL(), cSaveScr )

		DO CASE
		CASE nKey == K_UP
			IF ( nTop != 0 )
				nTop--
				nBottom--
			ENDIF
		CASE nKey == K_DOWN
			IF nBottom != MAXROW()
				nBottom++
				nTop++
			ENDIF
		CASE nKey == K_LEFT
			IF nLeft != 0
				nLeft--
				nRight--
			ENDIF
		CASE nKey == K_RIGHT
			IF nRight != MAXCOL()
				nRight++
				nLeft++
			ENDIF
		CASE nKey == K_ENTER
			EXIT
		ENDCASE
		@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	ENDDO
	SETCOLOR(cColor)

	oBrowse:nTop += (nTop - oBrowse:cargo[BOXTOP])
	oBrowse:nLeft += (nLeft - oBrowse:cargo[BOXLEFT])
	oBrowse:nBottom += (nBottom - oBrowse:cargo[BOXBOTTOM])
	oBrowse:nRight  += (nRight - oBrowse:cargo[BOXRIGHT])

	DISPBEGIN()
	// clear area occupied by old browse position
	@ oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT] CLEAR TO ;
	  oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT]

	oBrowse:cargo[BOXTOP] := nTop
	oBrowse:cargo[BOXLEFT] := nLeft
	oBrowse:cargo[BOXBOTTOM] := nBottom
	oBrowse:cargo[BOXRIGHT] := nRight

	DISPBOX(oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT],;
		oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT],;
		oBrowse:cargo[BOXTYPE]+" ", oBrowse:cargo[BOXBORDER])

	DISPEND()

	RETURN (NIL)

<ENDIF>
<IFKEY><SIZE>

//===========================================================================
/*
 SizeBrowse()

 Dynamically size the browse box.
 Use the arrow keys to size the box, <return> to select size.
 Updates browse object to reflect new box coordinates.

 SYNTAX:  SizeBrowse(expO1)

    <expB1> -  browse object we are sizing

 RETURNS: NIL
 --------------------------------------------------------------------------*/
FUNCTION SizeBrowse( oBrowse )
	LOCAL cSaveScr
	LOCAL cColor
	LOCAL nKey
	LOCAL nTop, nLeft, nBottom, nRight

	nTop := oBrowse:cargo[BOXTOP]
	nLeft := oBrowse:cargo[BOXLEFT]
	nBottom := oBrowse:cargo[BOXBOTTOM]
	nRight := oBrowse:cargo[BOXRIGHT]

	// save original screen
	cColor := SETCOLOR("W+/N")
	cSaveScr := SAVESCREEN( 0, 0, MAXROW(), MAXCOL() )

	@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	DO WHILE .T.
		nKey := INKEY(0)
		RESTSCREEN( 0, 0, MAXROW(), MAXCOL(), cSaveScr )

		DO CASE
		CASE nKey == K_UP
			nBottom := IF( nBottom > nTop+2, --nBottom, nBottom )
		CASE nKey == K_DOWN
			nBottom := IF( nBottom != MAXROW(), ++nBottom, nBottom )
		CASE nKey == K_LEFT
			nRight := IF( nRight > nLeft+2, --nRight, nRight )
		CASE nKey == K_RIGHT
			nRight := IF( nRight != MAXCOL(), ++nRight, nRight )
		CASE nKey == K_ENTER
			EXIT
		ENDCASE

		@ nTop, nLeft, nBottom, nRight BOX CHR(219)

	ENDDO
	SETCOLOR(cColor)

	DISPBEGIN()

	// clear area occupied by old browse position
	@ oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT] CLEAR TO ;
	  oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT]

	oBrowse:nBottom += (nBottom - oBrowse:cargo[BOXBOTTOM])
	oBrowse:nRight  += (nRight - oBrowse:cargo[BOXRIGHT])
	oBrowse:cargo[BOXBOTTOM] := nBottom
	oBrowse:cargo[BOXRIGHT] := nRight

	DISPBOX(oBrowse:cargo[BOXTOP], oBrowse:cargo[BOXLEFT],;
		oBrowse:cargo[BOXBOTTOM], oBrowse:cargo[BOXRIGHT],;
		oBrowse:cargo[BOXTYPE], oBrowse:cargo[BOXBORDER])

	DISPEND()

	RETURN (NIL)

<ENDIF>
